<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\Component\Utility\Html;
use Drupal\Core\Ajax\AjaxFormHelperTrait;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\RedirectCommand;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormFactoryInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\entity_template\BlueprintStorageInterface;
use Drupal\entity_template\BlueprintTempstoreRepository;
use Drupal\entity_template\TemplateManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

class AddTemplateForm extends FormBase {
  use AjaxFormHelperTrait;

  /**
   * @var \Drupal\entity_template\TemplateManager
   */
  protected $templateManager;

  /**
   * @var \Drupal\entity_template\BlueprintStorageInterface
   */
  protected $blueprintStorage;

  /**
   * @var \Drupal\Core\Plugin\PluginFormFactoryInterface
   */
  protected $pluginFormFactory;

  /**
   * @var \Drupal\entity_template\BlueprintTempstoreRepository
   */
  protected $blueprintTempstore;

  /**
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface
   */
  protected $plugin;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.template'),
      $container->get('plugin_form.factory'),
      $container->get('entity_template.blueprint_tempstore_repository')
    );
  }

  public function __construct(
    TemplateManager $template_manager,
    PluginFormFactoryInterface $plugin_form_factory,
    BlueprintTempstoreRepository $blueprint_tempstore_repository
  ) {
    $this->templateManager = $template_manager;
    $this->pluginFormFactory = $plugin_form_factory;
    $this->blueprintTempstore = $blueprint_tempstore_repository;
  }

  /**
   * Returns a unique string identifying the form.
   *
   * The returned ID should be a unique string that can be a valid PHP function
   * name, since it's used in hook implementation names such as
   * hook_form_FORM_ID_alter().
   *
   * @return string
   *   The unique string identifying the form.
   */
  public function getFormId() {
    return 'add_template_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(
    array $form,
    FormStateInterface $form_state,
    BlueprintStorageInterface $blueprint_storage = NULL,
    $plugin_id = ''
  ) {
    $this->blueprintStorage = $blueprint_storage;
    if (!$this->plugin) {
      $this->plugin = $this->templateManager->createInstance($plugin_id, [
        'blueprint' => $blueprint_storage,
      ]);
    }

    if (
      ($this->plugin instanceof PluginWithFormsInterface) &&
      $this->plugin->hasFormClass("add")
    ) {
      $plugin_form = $this->pluginFormFactory->createInstance(
        $this->plugin,
        "add"
      );

      $form = $plugin_form->buildConfigurationForm($form, $form_state);
    }

    $form['actions'] = [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Add Template'),
        '#button_type' => 'primary',
      ],
    ];
    if ($this->isAjax()) {
      $form['actions']['submit']['#ajax']['callback'] = '::ajaxSubmit';
      // @todo static::ajaxSubmit() requires data-drupal-selector to be the same
      //   between the various Ajax requests. A bug in
      //   \Drupal\Core\Form\FormBuilder prevents that from happening unless
      //   $form['#id'] is also the same. Normally, #id is set to a unique HTML
      //   ID via Html::getUniqueId(), but here we bypass that in order to work
      //   around the data-drupal-selector bug. This is okay so long as we
      //   assume that this form only ever occurs once on a page. Remove this
      //   workaround in https://www.drupal.org/node/2897377.
      $form['#id'] = Html::getId($form_state->getBuildInfo()['form_id']);
    }

    return $form;
  }

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    if (
      ($this->plugin instanceof PluginWithFormsInterface) &&
      $this->plugin->hasFormClass("add")
    ) {
      $plugin_form = $this->pluginFormFactory->createInstance(
        $this->plugin,
        "add"
      );

      $plugin_form->submitConfigurationForm($form, $form_state);
    }

    $this->blueprintStorage->addTemplate($this->plugin->getConfiguration());
    $this->blueprintTempstore->set($this->blueprintStorage);
  }

  /**
   * {@inheritdoc}
   */
  protected function successfulAjaxSubmit(array $form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $response->addCommand(new RedirectCommand($this->blueprintStorage
      ->getEditUrl()
      ->setOption('absolute', TRUE)
      ->toString()
    ));
    return $response;
  }
}
