<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\charts_highcharts\Settings\Highcharts\Pane;
use Drupal\Core\Entity\Display\EntityFormDisplayInterface;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Render\Element;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\entity_template\Exception\MultipleAvailableBlueprintException;
use Drupal\entity_template\Exception\NoAvailableBlueprintException;
use Drupal\entity_template\TemplateBuilderManager;
use Drupal\typed_data\Widget\FormWidgetManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * This form allows the user to create entities from an specific builder. This
 * form has 3 steps:
 *
 * 1. The builder selection. The form will show a table of the available
 * builders for the user to select one. The columns of the table are as follows:
 *   - A radio button for user to pick the builder.
 *   - The label of the builder.
 *   - The return type of the builder (hidden if the 'entity_type_id' parameter
 *     is available).
 *
 * As mentioned above, an 'entity_type_id' parameter can be passed in order to
 * filter the builders by return type.
 * The other optional parameter is 'builder_id', which will skip this first step
 * and go directly to the step two.
 *
 * 2. The parameters. If the builder (either selected by the user of given as a
 * parameter) has contexts that need to be entered, the form will show the
 * fields to input these values. The form relies on typed data form widgets to
 * show the correct input for each parameter, if the manager cannot a find an
 * applicable widget for the context, the broken widget will be shown, is up to
 * the developer to create custom widgets for unsupported typed data.
 *
 * If there are no parameters a button will be shown so user can continue with
 * the execution, this button is expected to be automatically submitted by a js
 * script so users don't have to do an extra click.
 *
 * After entering the parameters, the form will attempt to execute the builder,
 * if the execution is successful, the user will be redirected to the canonical
 * page of the last created entity if available, otherwise the page will be
 * reloaded.
 * Notice: A loop can be created here if the following conditions are met:
 *   - A 'builder_id' was passed to the form.
 *   - The builder doesn't have parameters.
 *   - The builder does have only one blueprint.
 *   - The last created entity doesn't have a canonical page.
 *
 * If the execution fails because of multiple blueprints, the user will be taken
 * to step number three.
 *
 * 3. Blueprint selection. A select input field with the available blueprints
 * for the selected builder is shown on the form. The user needs to pick one and
 * continue with the execution which will be the same as in step two.
 *
 * @package Drupal\entity_template_ui\Form
 */
class CreateEntityFromTemplateForm extends FormBase {

  /**
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $templateBuilderManager;

  /**
   * @var \Drupal\typed_data\Widget\FormWidgetManagerInterface
   */
  protected $formWidgetManager;

  /**
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected $typedDataManager;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The message to display upon successful submission. Return NULL to not show
   * any messages.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   */
  protected function getSuccessfulMessage() {
    return $this->t('Builder was successfully executed.');
  }

  /**
   * CreateEntityFromTemplateForm constructor.
   *
   * @param \Drupal\entity_template\TemplateBuilderManager $template_builder_manager
   * @param \Drupal\typed_data\Widget\FormWidgetManagerInterface $form_widget_manager
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typed_data_manager
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   */
  public function __construct(
    TemplateBuilderManager $template_builder_manager,
    FormWidgetManagerInterface $form_widget_manager,
    TypedDataManagerInterface $typed_data_manager,
    EntityTypeManagerInterface  $entity_type_manager
  ) {
    $this->templateBuilderManager = $template_builder_manager;
    $this->formWidgetManager = $form_widget_manager;
    $this->typedDataManager = $typed_data_manager;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.builder'),
      $container->get('plugin.manager.typed_data_form_widget'),
      $container->get('typed_data_manager'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'create_entity_from_template';
  }

  /**
   * Form constructor.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param string $entity_type_id
   * @param string $builder_id
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Exception
   */
  public function buildForm(array $form, FormStateInterface $form_state, $entity_type_id = NULL, $builder_id = NULL) {
    if (!$form_state->get('builders') && !$form_state->get('builder')) {
      if ($builder_id && $this->templateBuilderManager->hasDefinition($builder_id)) {
        $builder = $this->templateBuilderManager->createInstance($builder_id);

        if ($builder->getAvailableBlueprints()) {
          $form_state->set('builder', $builder);
          $form_state->set('step', 'parameters');
        }
      }

      if (!$form_state->get('builder')) {
        $builders = [];
        foreach ($this->templateBuilderManager->getDefinitions() as $id => $definition) {
          $builder = $this->templateBuilderManager->createInstance($id);

          // If the return type does not have the colon in it, it must be of
          // type entity or entity[]
          $return_type = $builder->getReturnType();
          $return_entity_type_id = '';
          if (strpos($return_type, ':') !== FALSE) {
            [, $return_entity_type_id] = explode(':', $return_type, 2);
          }

          if (!$entity_type_id || strpos($return_entity_type_id, $entity_type_id) === 0) {
            $builders[$id] = $builder;
          }
        }

        $form_state->set('entity_type_id', $entity_type_id);
        $form_state->set('builders', $builders);
        $form_state->set('step', 'select_builder');
      }
    }

    if ($form_state->get('step') == 'select_builder') {
      $form = $this->buildSelectBuilderForm($form, $form_state);
    }
    else if ($form_state->get('step') == 'parameters') {
      $form = $this->buildParametersForm($form, $form_state);
    }
    else if ($form_state->get('step') === 'blueprints') {
      $form = $this->buildBlueprintsForm($form, $form_state);
    }
    else if ($form_state->get('step') === 'entities') {
      $form = $this->buildEntitiesForm($form, $form_state);
    }

    return $form;
  }

  /**
   * Build the select parameters form.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  protected function buildSelectBuilderForm(array $form, FormStateInterface $form_state) {
    $entity_type_id = $form_state->get('entity_type_id');
    $builders = $form_state->get('builders');
    $headers = [
      '',
      $this->t('Template'),
    ];
    if (!$entity_type_id) {
      $headers[] = $this->t('Builds');
    }
    $form['builders'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Select a Template'),
      'table' => [
        '#type' => 'table',
        '#header' => $headers,
        '#empty' => $this->t('No results.'),
      ],
    ];

    $rows = [];
    foreach ($builders as $key => $builder) {
      $row = [
        'select' => [
          '#type' => 'radio',
          '#title' => $this->t('Select'),
          '#title_display' => 'invisible',
          '#parents' => ['builder'],
          '#return_value' => $key,
        ],
        'label' => [
          '#plain_text' => $builder->getPluginDefinition()['label'],
        ],
      ];
      if (!$entity_type_id) {
        $row['builds'] = ['#plain_text' => $builder->getReturnTypeLabel()];
      }
      $rows[] = $row;
    }

    $form['builders']['table'] += $rows;
    $form['select_builder'] = [
      '#type' => 'submit',
      '#value' => $this->t('Select'),
      '#validate' => ['::validateFormSelectBuilder'],
      '#submit' => ['::submitFormSelectBuilder'],
    ];

    return $form;
  }

  /**
   * Builds the parameters form step.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   * @throws \Exception
   */
  protected function buildParametersForm(array $form, FormStateInterface $form_state) {
    $builder = $form_state->get('builder');
    $parameters = $builder->getPluginDefinition()['context_definitions'];
    $route_match = $this->getRouteMatch();

    if (!empty($parameters)) {
      $provided_params = [];
      foreach ($parameters as $id => $context_definition) {
        if ($param = $route_match->getParameter($id)) {
          $provided_params[$id] = $param;
        }
        else {
          /** @var \Drupal\Core\Plugin\Context\ContextDefinition $context_definition */
          $data_definition = $context_definition->getDataDefinition();
          $typed_data = $this->typedDataManager->createInstance(
            $context_definition->getDataType(),
            [
              'data_definition' => $data_definition,
              'name' => $id,
              'parent' => NULL,
            ]
          );

          /** @var \Drupal\typed_data\Widget\FormWidgetInterface $widget */
          $widget = $this->formWidgetManager->getFormWidget($data_definition);
          $form['params'][$id] = [
            '#widget' => $widget,
            '#typed_data' => $typed_data,
          ];
          $form['params'][$id] += $widget->form(
            $typed_data,
            SubformState::createForSubform($form['params'][$id], $form, $form_state)
          );
        }
      }

      $form_state->set('params', $provided_params);
    }

    if (!empty($parameters) && count($provided_params) < count($parameters)) {
      $form['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
      ];
    }
    else {
      $form['no_params_message'] = [
        '#plain_text' => $this->t('Click on submit if it is not running already.')
      ];
      $form['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
        '#attributes' => [
          'class' => ['entity-template-ui-trigger-submit']
        ],
      ];
      $form['#attached']['library'][] = 'entity_template_ui/entity_template_ui.trigger_submit';
    }

    return $form;
  }

  /**
   * Build the blueprints form step.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  protected function buildBlueprintsForm(array $form, FormStateInterface $form_state) {
    $builder = $form_state->get('builder');
    $blueprints = $builder->getAvailableBlueprints();
    $blueprint_options = [];
    foreach ($blueprints as $key => $blueprint) {
      $blueprint_options[$key] = $blueprint->getEntity()->label();
    }

    $form['blueprint_key'] = [
      '#type' => 'select',
      '#title' => $this->t('Select the blueprint'),
      '#options' => $blueprint_options,
      '#required' => TRUE,
    ];

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
      '#submit' => [
        '::submitFormSelectBlueprint',
        '::submitForm',
      ],
    ];

    return $form;
  }

  /**
   * Build the entities form.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  protected function buildEntitiesForm(array $form, FormStateInterface $form_state) {
    $entity = $form_state->get('current_entity');
    $form['_result_summary'] = [
      '#type' => 'inline_template',
      '#template' => 'Editing {{ current }} of {{ total }}',
      '#context' => [
        'current' => array_search($entity->uuid(), array_keys($form_state->get('entities'))) + 1,
        'total' => count($form_state->get('entities')),
      ],
    ];

    if ($display = $this->getCurrentEntityFormDisplay($form_state, $entity)) {
      $display->buildForm($entity, $form, $form_state);
    }

    $form['actions'] = $this->buildEntitiesFormActions($entity, $form, $form_state);

    return $form;
  }

  protected function getCurrentEntityFormDisplay(FormStateInterface $form_state, EntityInterface $entity = NULL):? EntityFormDisplayInterface {
    if (!$form_state->get('current_entity_form_display')) {
      if (!$entity || !($entity instanceof FieldableEntityInterface)) {
        return NULL;
      }

      $form_state->set('current_entity_form_display', EntityFormDisplay::collectRenderDisplay($entity, 'add'));
    }

    return $form_state->get('current_entity_form_display');
  }

  /**
   * Build the actions.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  protected function buildEntitiesFormActions(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    return [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Save'),
        '#submit' => [
          '::submitEntitiesForm',
          '::submitEntitiesFormNextEntity',
        ],
        '#validate' => [
          '::validateEntitiesForm',
        ],
      ],
    ];
  }

  /**
   * Validate the entities form.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function validateEntitiesForm(array $form, FormStateInterface $form_state) {
    $entity = clone $form_state->get('current_entity');

    if ($display = $this->getCurrentEntityFormDisplay($form_state)) {
      $display->extractFormValues($entity, $form, $form_state);
    }
    $entity->setValidationRequired(!$form_state->getTemporaryValue('current_entity_validated'));

    $violations = $entity->validate();

    // Remove violations of inaccessible fields.
    $violations->filterByFieldAccess($this->currentUser());

    // Flag entity level violations.
    foreach ($violations->getEntityViolations() as $violation) {
      /** @var \Symfony\Component\Validator\ConstraintViolationInterface $violation */
      $form_state->setErrorByName(str_replace('.', '][', $violation->getPropertyPath()), $violation->getMessage());
    }

    // Let the form display flag violations of its fields.
    if ($display) {
      $display->flagWidgetsErrorsFromViolations($violations, $form, $form_state);
    }

    // The entity was validated.
    $entity->setValidationRequired(FALSE);
    $form_state->setTemporaryValue('current_entity_validated', TRUE);
  }

  /**
   * Submit the entities form.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitEntitiesForm(array $form, FormStateInterface $form_state) {
    $entity = $form_state->get('current_entity');
    if ($display = $this->getCurrentEntityFormDisplay($form_state)) {
      $display->extractFormValues($entity, $form, $form_state);
    }
    $entity->save();
  }

  /**
   * Submit the entities form and go to the next entity.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitEntitiesFormNextEntity(array $form, FormStateInterface $form_state) {
    $current_entity = $form_state->get('current_entity');
    $entities = $form_state->get('entities');
    $keys = array_keys($entities);
    if (array_search($current_entity->uuid(), $keys) + 1 < count($keys)) {
      $next_key = array_search($current_entity->uuid(), $keys) + 1;
      $form_state->set('current_entity', $entities[$next_key]);
      $form_state->setRebuild(TRUE);
    }
    else {
      if (!$form_state->getRedirect() && $current_entity->hasLinkTemplate('canonical')) {
        $form_state->setRedirect('entity.'.$current_entity->getEntityTypeId().'.canonical', [
          $current_entity->getEntityTypeId() => $current_entity->id(),
        ]);
      }
    }
  }

  /**
   * Validate callback for selecting a builder.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function validateFormSelectBuilder(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getValue('builder')) {
      $form_state->setError($form['builders'], $this->t('Select a builder.'));
    }
  }

  /**
   * Submit callback for selecting a builder.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitFormSelectBuilder(array &$form, FormStateInterface $form_state) {
    $builder = $form_state->get(['builders', $form_state->getValue('builder')]);

    $form_state->set('builder', $builder);
    $form_state->set('step', 'parameters');
    $form_state->setRebuild();
  }

  /**
   * Submit callback for selecting a blueprint.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitFormSelectBlueprint(array &$form, FormStateInterface $form_state) {
    $form_state->set('blueprint_key', $form_state->getValue('blueprint_key'));
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    /** @var \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderBase $builder */
    $builder = $form_state->get('builder');

    try {
      // Store the params in the form state to be used again if a blueprint
      // needs to be selected.
      if (!empty($form['params'])) {
        $parameters = $form_state->get('params') ?: [];
        foreach (Element::children($form['params']) as $id) {
          /** @var \Drupal\typed_data\Widget\FormWidgetInterface $widget */
          $widget = $form['params'][$id]['#widget'];
          /** @var \Drupal\Core\TypedData\TypedDataInterface $typed_data */
          $typed_data = $form['params'][$id]['#typed_data'];

          $widget->extractFormValues(
            $typed_data,
            SubformState::createForSubform($form['params'][$id], $form, $form_state)
          );
          $parameters[$id] = $typed_data->getValue();
        }

        $form_state->set('params', $parameters);
      }

      /** @var \Drupal\entity_template\BlueprintResult $result */
      $params = $form_state->get('params') ?: [];
      foreach ($builder->getContextDefinitions() as $id => $definition) {
        // Upcast the entity ids that might be in params.
        if (isset($params[$id]) && substr($definition->getDataType(), 0, 7) === 'entity:' && !is_object($params[$id])) {
          [,$entity_type_id] = explode(':', $definition->getDataType());
          $params[$id] = $this->entityTypeManager->getStorage($entity_type_id)->load($params[$id]);
        }
      }
      $result = $builder->execute($params, NULL, $form_state->get('blueprint_key'));
      $items = $result->getItems();
      if (!empty($items)) {
        $form_state->set('entities', $items);
        $form_state->set('current_entity', reset($items));
        $form_state->set('step', 'entities');
        $form_state->setRebuild(TRUE);

        if ($message = $this->getSuccessfulMessage()) {
          $this->messenger()->addStatus($message);
        }
      }
      else {
        $this->messenger()->addWarning(new TranslatableMarkup('No items were created.'));
      }
    }
    catch (MultipleAvailableBlueprintException $e) {
      $form_state->set('step', 'blueprints');
      $form_state->setRebuild();
    }
    catch (NoAvailableBlueprintException $e) {
      $this->messenger()->addWarning('The selected builder does not have any available blueprints.');
      $form_state->set('builder', FALSE);
      $form_state->set('step', 'select_builder');
      $form_state->setRebuild();
    }
  }

}
