<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\Core\Ajax\AjaxFormHelperTrait;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface;
use Drupal\entity_template\BlueprintStorageInterface;
use Drupal\entity_template\BlueprintTempstoreRepository;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class RemoveComponentForm
 *
 * @package Drupal\entity_template_ui\Form
 */
class RemoveComponentForm extends ConfirmFormBase {
  use AjaxFormHelperTrait;
  use TemplateAjaxFormHelperTrait;

  /**
   * @var \Drupal\entity_template\BlueprintStorageInterface
   */
  protected $blueprintStorage;

  /**
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface
   */
  protected $template;

  /**
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface
   */
  protected $rootTemplate;

  /**
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Component\ComponentInterface
   */
  protected $component;

  /**
   * @var \Drupal\entity_template\BlueprintTempstoreRepository
   */
  protected $blueprintTempstoreRepository;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_template.blueprint_tempstore_repository')
    );
  }

  /**
   * RemoveComponentForm constructor.
   *
   * @param \Drupal\entity_template\BlueprintTempstoreRepository $blueprint_tempstore_repository
   */
  public function __construct(BlueprintTempstoreRepository $blueprint_tempstore_repository) {
    $this->blueprintTempstoreRepository = $blueprint_tempstore_repository;
  }

  /**
   * Returns the question to ask the user.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The form question. The page title will be set to this value.
   */
  public function getQuestion() {
    return new TranslatableMarkup(
      'Are you sure you want to remove @component from @template?',
      [
        '@template' => $this->template->label(),
        '@component' => $this->component->label()
      ]
    );
  }

  /**
   * Returns the route to go to if the user cancels the action.
   *
   * @return \Drupal\Core\Url
   *   A URL object.
   */
  public function getCancelUrl() {
    return $this->blueprintStorage->getEditTemplateUrl($this->rootTemplate->getKey());
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(
    array $form,
    FormStateInterface $form_state,
    BlueprintStorageInterface $blueprint_storage = NULL,
    $template_key = '',
    $sub_template = '',
    $component_id = ''
  ) {
    $this->blueprintStorage = $blueprint_storage;
    $this->rootTemplate = $this->template = $this->blueprintStorage->getTemplate($template_key);

    if (!empty($sub_template)) {
      $path = explode(':', $sub_template);
      while ($bit = array_shift($path)) {
        $this->template = $this->template->getComponent($bit);

        if (!$this->template || !($this->template instanceof  TemplateInterface)) {
          throw new NotFoundHttpException();
        }
      }
    }

    $this->component = $this->template->getComponent($component_id);

    $form = parent::buildForm($form, $form_state);

    if ($this->isAjax()) {
      $form['actions']['submit']['#ajax']['callback'] = '::ajaxSubmit';
      $form['actions']['cancel']['#attributes']['class'][] = 'dialog-cancel';
    }

    return $form;
  }

  /**
   * Returns a unique string identifying the form.
   *
   * The returned ID should be a unique string that can be a valid PHP function
   * name, since it's used in hook implementation names such as
   * hook_form_FORM_ID_alter().
   *
   * @return string
   *   The unique string identifying the form.
   */
  public function getFormId() {
    return 'remove_component_form';
  }

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->template->removeComponent($this->component->getConfiguration()['uuid']);
    $this->blueprintTempstoreRepository->set($this->blueprintStorage);

    $form_state->setRedirectUrl(
      $this->blueprintStorage->getEditTemplateUrl($this->rootTemplate->getKey())
    );
  }
}
