<?php

namespace Drupal\entity_template_ui\PluginForm;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormBase;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class TemplateAddForm
 *
 * @package Drupal\entity_template_ui\PluginForm
 */
abstract class TemplateAddForm extends PluginFormBase implements ContainerInjectionInterface {

  /**
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $entityTypeBundleInfo;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Template\BlueprintTemplateInterface
   */
  protected $plugin;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * TemplateAddForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $entity_type_bundle_info
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfo $entity_type_bundle_info
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * Get the required return type.
   *
   * @return string
   */
  protected abstract function getRequiredReturnType();

  /**
   * Form constructor.
   *
   * Plugin forms are embedded in other forms. In order to know where the plugin
   * form is located in the parent form, #parents and #array_parents must be
   * known, but these are not available during the initial build phase. In order
   * to have these properties available when building the plugin form's
   * elements, let this method return a form element that has a #process
   * callback and build the rest of the form in the callback. By the time the
   * callback is executed, the element's #parents and #array_parents properties
   * will have been set by the form API. For more documentation on #parents and
   * #array_parents, see \Drupal\Core\Render\Element\FormElement.
   *
   * @param array $form
   *   An associative array containing the initial structure of the plugin form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form. Calling code should pass on a subform
   *   state created through
   *   \Drupal\Core\Form\SubformState::createForSubform().
   *
   * @return array
   *   The form structure.
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['label'] = [
      '#type' => 'textfield',
      '#title' => new TranslatableMarkup('Label'),
      '#description' => new TranslatableMarkup(
        'A label to help administrators view this template.'
      ),
      '#default_value' => $this->plugin->label(),
    ];

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => new TranslatableMarkup('Description'),
      '#description' => new TranslatableMarkup('A short description of this template. This is only visible to administrators.'),
      '#size' => 3,
    ];

    $required_return_type = $this->getRequiredReturnType();
    if (substr_count($required_return_type, ':') === 0) {
      $options = [];
      foreach ($this->entityTypeManager->getDefinitions() as $id => $definition) {
        if (!$definition->hasKey('bundle')) {
          $options[$id] = $definition->getLabel();
        }
        else {
          $rendered_label = $definition->getLabel()->render();
          $options[$rendered_label] = [];
          foreach ($this->entityTypeBundleInfo->getBundleInfo($id) as $bundle => $binfo) {
            $options[$rendered_label]["{$id}:{$bundle}"] = $binfo['label'];
          }
        }
      }

      $form['target_entity_type_id_bundle'] = [
        '#type' => 'select',
        '#options' => $options,
        '#title' => new TranslatableMarkup('Return Type'),
        '#description' => new TranslatableMarkup('What sort of entity does this template create?'),
        '#default_value' => $this->plugin->getConfiguration()['target_entity_type_id'],
        '#element_validate' => [
          [static::class, 'combinedEntityTypeIdBundleElementValidate']
        ]
      ];
    }
    else if (substr_count($required_return_type, ':') === 1) {
      list(,$entity_type_id) = explode(':', $required_return_type);
      if (substr($entity_type_id, -2) === '[]') {
        $entity_type_id = substr($entity_type_id, 0, -2);
      }

      $options = [];
      foreach ($this->entityTypeBundleInfo->getBundleInfo($entity_type_id) as $bundle => $binfo) {
        $options[$bundle] = $binfo['label'];
      }

      $form['target_entity_bundle'] = [
        '#type' => 'select',
        '#options' => $options,
        '#title' => new TranslatableMarkup('Return Type'),
        '#description' => new TranslatableMarkup('What sort of entity does this template create?'),
        '#default_value' => $this->plugin->getConfiguration()['target_entity_bundle'],
      ];
    }

    return $form;
  }

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the plugin form as built
   *   by static::buildConfigurationForm().
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form. Calling code should pass on a subform
   *   state created through
   *   \Drupal\Core\Form\SubformState::createForSubform().
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $configuration = $this->plugin->getConfiguration();
    $configuration['label'] = $form_state->getValue('label');
    $configuration['description'] = $form_state->getValue('description');

    if ($entity_type_id = $form_state->getValue('target_entity_type_id')) {
      $configuration['target_entity_type_id'] = $entity_type_id;
    }
    if ($bundle = $form_state->getValue('target_entity_bundle')) {
      $configuration['target_entity_bundle'] = $bundle;
    }

    $this->plugin->setConfiguration($configuration);
  }

  /**
   * Validate the combined entity type and bundle element.
   *
   * @param array $element
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public static function combinedEntityTypeIdBundleElementValidate(
    array $element,
    FormStateInterface $form_state
  ) {
    $parents = $element['#parents'];
    $value = $form_state->getValue($parents);
    array_pop($parents);

    $entity_type_parents = array_merge($parents, ['target_entity_type_id']);
    $bundle_parents = array_merge($parents, ['target_entity_bundle']);

    if (strpos($value, ':')) {
      list($entity_type_id, $bundle) = explode(':', $value);
      $form_state->setValue($entity_type_parents, $entity_type_id);
      $form_state->setValue($bundle_parents, $bundle);
    }
    else {

      $form_state->setValue($entity_type_parents, $value);
    }
  }
}
