<?php

namespace Drupal\entity_template\Form;

use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\TypedData\EntityDataDefinition;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\entity_template\EntityBuilderTempStoreRepository;
use Drupal\entity_template\TemplateBuilderManager;
use Http\Discovery\Exception\NotFoundException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Session\SessionInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Builder parameter provision form.
 */
class BuilderParameterProvisionForm extends FormBase {

  /**
   * The builder name.
   *
   * @var string
   */
  protected $builderName;

  /**
   * The session key.
   *
   * @var string
   */
  protected $sessionKey;

  /**
   * The template builder manager.
   *
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $templateBuilderManager;

  /**
   * The entity builder tempstore.
   *
   * @var \Drupal\entity_template\EntityBuilderTempStoreRepository
   */
  protected $entityBuilderTempStore;

  /**
   * The session.
   *
   * @var \Symfony\Component\HttpFoundation\Session\Session
   */
  protected $session;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new BuilderParameterProvisionForm.
   *
   * @param \Drupal\entity_template\TemplateBuilderManager $template_builder_manager
   *   The template builder manager.
   * @param \Drupal\entity_template\EntityBuilderTempStoreRepository $entity_builder_tempstore
   *   The entity builder tempstore.
   * @param \Symfony\Component\HttpFoundation\Session\SessionInterface $session
   *   The session.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The UUID service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(TemplateBuilderManager $template_builder_manager, EntityBuilderTempStoreRepository $entity_builder_tempstore, SessionInterface $session, UuidInterface $uuid_service, EntityTypeManagerInterface $entity_type_manager) {
    $this->templateBuilderManager = $template_builder_manager;
    $this->entityBuilderTempStore = $entity_builder_tempstore;
    $this->session = $session;
    $this->uuidService = $uuid_service;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.builder'),
      $container->get('entity_template.entity_builder_tempstore'),
      $container->get('session'),
      $container->get('uuid'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'builder_parameter_provision_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $builder_name = NULL) {
    if (!$builder_name) {
      throw new NotFoundHttpException();
    }
    $this->builderName = $builder_name;

    $this->sessionKey = $this->uuidService->generate();
    $this->session->set('entity_builder_' . $builder_name . '_session_key', $this->sessionKey);

    $builder = $this->templateBuilderManager->createInstance($this->builderName);

    foreach ($builder->getContextDefinitions() as $name => $definition) {
      $data_definition = $definition->getDataDefinition();
      // @todo CKAPPS-3696 typed_data module provides a plugin type for setting the values of data_definitions
      // @todo using a plugin means we can have different behaviours for different data types
      if ($data_definition instanceof EntityDataDefinition) {
        $form[$name] = [
          '#type' => 'entity_autocomplete',
          '#title' => new TranslatableMarkup(
            '@label',
            [
              '@label' => $definition->getLabel(),
            ]
          ),
          '#target_type' => str_replace('entity:', '', $definition->getDataType()),
        ];
      }
      else {
        $form[$name] = [
          '#type' => 'textfield',
          '#title' => new TranslatableMarkup(
            '@label',
            [
              '@label' => $definition->getLabel(),
            ]
          ),
        ];
      }

      if ($definition->isRequired()) {
        $form[$name]['#required'] = TRUE;
      }
    }

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => new TranslatableMarkup('Set Parameters'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $builder = $this->templateBuilderManager->createInstance($this->builderName);

    $builder_parameters = [];
    foreach ($builder->getContextDefinitions() as $name => $definition) {
      if (isset($form_state->getCompleteForm()[$name]['#target_type']) && ($definition instanceof EntityContextDefinition)) {
        $builder_parameters[$name] = $this->entityTypeManager->getStorage($form_state->getCompleteForm()[$name]['#target_type'])->load($form_state->getValue($name));
      } else {
        $builder_parameters[$name] = $form_state->getValue($name);
      }
    }

    $params = [
      'builder_name' => $this->builderName,
      'session_key' => $this->sessionKey,
    ];

    $this->entityBuilderTempStore->getTempstore()->set('builder_parameters__' . $this->sessionKey, $builder_parameters);

    $form_state->setRedirectUrl(Url::fromRoute($this->getTemplateBlueprintSelectionRoute(), $params));
  }

  /**
   * Gets the template blueprint selection route.
   *
   * @return string
   *   The route.
   */
  protected function getTemplateBlueprintSelectionRoute() {
    return 'entity_template.select_template_blueprint';
  }

}
