<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Builder;

use Drupal\Component\Plugin\Exception\ContextException;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\ContextAwarePluginTrait;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\entity_template\BlueprintStorageInterface;
use Drupal\entity_template\Exception\MultipleAvailableBlueprintException;
use Drupal\entity_template\BlueprintInterface;
use Drupal\entity_template\Exception\NoAvailableBlueprintException;
use Drupal\entity_template\TemplateBlueprintProviderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class BuilderBase
 *
 * @package Drupal\entity_template\Plugin\EntityTemplate\Builder
 */
abstract class BuilderBase extends PluginBase implements ContextAwarePluginInterface, BuilderInterface, ContainerFactoryPluginInterface {
  use ContextAwarePluginTrait;

  /**
   * @var \Drupal\entity_template\TemplateBlueprintProviderManager
   */
  protected $blueprintProviderManager;

  /**
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.entity_template.blueprint_provider'),
      $container->get('current_user'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * ConfigTemplateBuilder constructor.
   *
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    TemplateBlueprintProviderManager $blueprint_provider_manager,
    AccountInterface $current_user,
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfoInterface $entity_type_bundle_info
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->currentUser = $current_user;
    $this->blueprintProviderManager = $blueprint_provider_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public function returnsList() {
    return substr($this->getReturnType(), -2) === '[]';
  }

  /**
   * {@inheritdoc}
   */
  public function getReturnType() {
    return isset($this->pluginDefinition['return_type']) ? $this->pluginDefinition['return_type'] : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailableBlueprints(array $parameters = [], AccountInterface $account = NULL) {
    if (!$account) {
      $account = $this->currentUser;
    }

    $blueprints = [];
    foreach ($this->blueprintProviderManager->getDefinitions() as $provider_id => $definition) {
      /** @var \Drupal\entity_template\Plugin\EntityTemplate\BlueprintProvider\BlueprintProviderInterface $provider */
      $provider = $this->blueprintProviderManager->createInstance($provider_id);

      foreach ($provider->getAvailableBlueprints($this, $parameters, $account) as $key => $available_blueprint) {
        $blueprints[$provider_id.':'.$key] = $available_blueprint;
      }
    }

    uasort($blueprints, function(BlueprintInterface $a, BlueprintInterface $b) {
      if ($a->getPriority() === $b->getPriority()) {
        return 0;
      }

      return ($a->getPriority() > $b->getPriority()) ? -1 : 1;
    });

    return $blueprints;
  }

  /**
   * {@inheritdoc}
   */
  public function execute(array $parameters = [], AccountInterface $account = NULL, $blueprint_key = NULL) {
    if ($blueprint_key) {
      list($provider_id, $provider_key) = explode(':', $blueprint_key);

      /** @var \Drupal\entity_template\Plugin\EntityTemplate\BlueprintProvider\BlueprintProviderInterface $provider */
      $provider = $this->blueprintProviderManager->createInstance($provider_id);
      $available = $provider->getAvailableBlueprints($this, $parameters, $account);

      if (!isset($available[$provider_key])) {
        throw new \InvalidArgumentException('No blueprint available with key '.$blueprint_key);
      }

      $blueprint = $available[$provider_key];
    }
    else {
      $available = $this->getAvailableBlueprints($parameters, $account);

      if (empty($available)) {
        throw new NoAvailableBlueprintException('No blueprints are available for '.$this->getPluginId());
      }

      $best_key = key($available);
      /** @var \Drupal\entity_template\BlueprintInterface $blueprint */
      $blueprint = array_shift($available);
      if (!empty($available)) {
        if (reset($available)->getPriority() == $blueprint->getPriority()) {
          $blueprints = [
            $best_key => $blueprint,
          ];
          foreach ($available as $key => $available_blueprint) {
            if ($available_blueprint->getPriority() < $blueprint->getPriority()) {
              break;
            }

            $blueprints[$key] = $available_blueprint;
          }

          throw new MultipleAvailableBlueprintException("Unable to select best blueprint.", $blueprints);
        }
      }
    }

    if ($blueprint->getContextDefinition('current_user')) {
      $blueprint->setContext(
        'current_user',
        new Context(
          $blueprint->getContextDefinition('current_user'),
          $this->entityTypeManager->getStorage('user')->load($account ? $account->id() : $this->currentUser->id())
        )
      );
    }

    return $blueprint->execute($parameters);
  }

  /**
   * {@inheritdoc}
   */
  public function getAdminOperations() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function hasContextDefinition($name) {
    try {
      $this->getContextDefinition($name);
    }
    catch (ContextException $e) {
      return FALSE;
    }
    catch (PluginException $e) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get the return type label.
   *
   * @return string|\Drupal\Component\Render\MarkupInterface
   */
  public function getReturnTypeLabel() {
    $return_type = $this->getReturnType();
    $is_list = FALSE;

    if ($this->returnsList()) {
      $return_type = substr($return_type, 0, -2);
      $is_list = TRUE;
    }

    if ($return_type === 'entity') {
      return $is_list ? $this->t('List of Entities') : $this->t('Any Entity');
    }
    else if (substr_count($return_type, ':') === 1) {
      list(,$entity_type_id) = explode(':', $return_type);

      $definition = $this->entityTypeManager->getDefinition($entity_type_id);
      return $is_list ? $this->t('List of @label', ['@label' => $definition->getPluralLabel()]) : $definition->getLabel();
    }
    else {
      list (,$entity_type_id, $bundle) = explode(':', $return_type);

      $definition = $this->entityTypeManager->getDefinition($entity_type_id);
      $bundle_info = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id)[$bundle];

      return $is_list ?
        $this->t('List of @bundle @label', [
          '@label' => $definition->getPluralLabel(),
          '@bundle' => $bundle_info['label'],
        ]) :
        $this->t('@bundle @label', [
          '@bundle' => $bundle_info['label'],
          '@label' => $definition->getLabel()
        ]);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultBlueprintEditTemplateUrl(
    BlueprintStorageInterface $blueprint_storage,
    string $key
  ): Url {
    return Url::fromRoute(
      'entity_template_ui.edit_template',
      [
        'template_blueprint_provider' => $blueprint_storage->getProvider()->getPluginId(),
        'blueprint_storage' => $blueprint_storage->getProviderKey(),
        'template_id' => $key,
      ]
    );
  }

}
