<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Component;

use Drupal\Core\TypedData\Exception\MissingDataException;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface;
use Drupal\typed_data\DataFetcherInterface;

/**
 * Trait to help with data selection.
 *
 * @package Drupal\entity_template\Plugin\EntityTemplate\Component
 */
trait DataSelectComponentTrait {

  /**
   * The data fetcher service.
   *
   * @var \Drupal\typed_data\DataFetcherInterface
   */
  protected $dataFetcher;

  /**
   * Select the correct typed data.
   *
   * @param string $selector
   *   The selector string.
   *
   * @return \Drupal\Core\TypedData\TypedDataInterface
   *   The selected and filtered data.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   *   If data is missing.
   */
  protected function selectData(string $selector) : TypedDataInterface {
    // The apply filters method is currently in an MR on the typed data module.
    // See https://www.drupal.org/project/typed_data/issues/3244608
    if (is_callable([$this->dataFetcher, 'applyFilters'])) {
      [$paths, $filters] = $this->getDataFetcher()->parsePropertyPathAndFilters($selector);
    }
    else {
      $filters = [];
      $paths = explode('.', $selector);
    }
    $context = array_shift($paths);

    if (empty($paths)) {
      $data = $this->getContextProvidingTemplate()->getContext($context)->getContextData();
    }
    else {
      $data = $this->getDataFetcher()->fetchDataBySubPaths(
        $this->getContextProvidingTemplate()->getContext($context)->getContextData(),
        $paths
      );
    }

    if ($filters) {
      $data = $this->getDataFetcher()->applyFilters($data, $filters);
    }

    return $data;
  }

  /**
   * Get the data fetcher service.
   *
   * @return \Drupal\typed_data\DataFetcherInterface
   *   The data fetcher service.
   */
  protected function getDataFetcher() : DataFetcherInterface {
    if (!$this->dataFetcher) {
      $this->dataFetcher = \Drupal::service('typed_data.data_fetcher');
    }

    return $this->dataFetcher;
  }

  /**
   * Get the context providing template.
   *
   * @return \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface
   *   The context providing template.
   */
  public abstract function getContextProvidingTemplate() : TemplateInterface;

}
